@@ -18,13 +18,20 @@ import logging
 import secrets
 import time
 from typing import Dict, Text
-from bs4 import BeautifulSoup
-import yarl
 
 import aiohttp
+from bs4 import BeautifulSoup
+import httpx
+import yarl
 from yarl import URL
 
-from teslajsonpy.const import DRIVING_INTERVAL, WEBSOCKET_TIMEOUT
+from teslajsonpy.const import (
+    API_URL,
+    AUTH_DOMAIN,
+    DRIVING_INTERVAL,
+    WEBSOCKET_TIMEOUT,
+    WS_URL,
+)
 from teslajsonpy.exceptions import IncompleteCredentials, TeslaException
 
 _LOGGER = logging.getLogger(__name__)
@@ -35,7 +42,7 @@ class Connection:
 
     def __init__(
         self,
-        websession: aiohttp.ClientSession,
+        websession: httpx.AsyncClient,
         email: Text = None,
         password: Text = None,
         access_token: Text = None,
@@ -51,8 +58,8 @@ class Connection:
         self.client_secret: Text = (
             "c7257eb71a564034f9419ee651c7d0e5f7" "aa6bfbd18bafb5c5c033b093bb2fa3"
         )
-        self.baseurl: Text = "https://owner-api.teslamotors.com"
-        self.websocket_url: Text = "wss://streaming.vn.teslamotors.com/streaming"
+        self.baseurl: Text = API_URL
+        self.websocket_url: Text = WS_URL
         self.api: Text = "/api/1/"
         self.expiration: int = expiration
         self.access_token = access_token
@@ -75,6 +82,8 @@ class Connection:
             self.__sethead(access_token=self.access_token, expiration=self.expiration)
             _LOGGER.debug("Connecting with existing access token")
         self.websocket = None
+        self.mfa_code: Text = ""
+        self.auth_domain: URL = URL(AUTH_DOMAIN)
 
     async def get(self, command):
         """Get data from API."""
@@ -101,7 +110,7 @@ class Connection:
                 if self.email and self.password:
                     _LOGGER.debug("Getting sso auth code using credentials")
                     self.code = await self.get_authorization_code(
-                        self.email, self.password
+                        self.email, self.password, mfa_code=self.mfa_code
                     )
                 else:
                     _LOGGER.debug("Using existing authorization code")
@@ -114,10 +123,10 @@ class Connection:
                     refresh_token=self.sso_oauth.get("refresh_token")
                 )
             if auth and all(
-                [
+                (
                     auth.get(item)
                     for item in ["access_token", "refresh_token", "expires_in"]
-                ]
+                )
             ):
                 self.sso_oauth = {
                     "access_token": auth["access_token"],
@@ -185,18 +194,23 @@ class Connection:
         _LOGGER.debug("%s: %s %s", method, url, data)
 
         try:
-            resp = await getattr(self.websession, method)(
-                url, data=data, headers=headers, cookies=cookies
-            )
-            data = await resp.json()
-            _LOGGER.debug("%s: %s", resp.status, json.dumps(data))
-            if resp.status > 299:
-                if resp.status == 401:
+            if data:
+                resp = await getattr(self.websession, method)(
+                    str(url), data=data, headers=headers, cookies=cookies
+                )
+            else:
+                resp = await getattr(self.websession, method)(
+                    str(url), headers=headers, cookies=cookies
+                )
+            data = resp.json()
+            _LOGGER.debug("%s: %s", resp.status_code, json.dumps(data))
+            if resp.status_code > 299:
+                if resp.status_code == 401:
                     if data and data.get("error") == "invalid_token":
-                        raise TeslaException(resp.status, "invalid_token")
-                elif resp.status == 408:
-                    raise TeslaException(resp.status, "vehicle_unavailable")
-                raise TeslaException(resp.status)
+                        raise TeslaException(resp.status_code, "invalid_token")
+                elif resp.status_code == 408:
+                    raise TeslaException(resp.status_code, "vehicle_unavailable")
+                raise TeslaException(resp.status_code)
             if data.get("error"):
                 # known errors:
                 #     'vehicle unavailable: {:error=>"vehicle unavailable:"}',
@@ -208,8 +222,8 @@ class Connection:
                 raise TeslaException(
                     f'{data.get("error")}:{data.get("error_description")}'
                 )
-        except aiohttp.ClientResponseError as exception_:
-            raise TeslaException(exception_.status) from exception_
+        except httpx.HTTPStatusError as exception_:
+            raise TeslaException(exception_.request.status_code) from exception_
         return data
 
     async def websocket_connect(self, vin: int, vehicle_id: int, **kwargs):
@@ -367,6 +381,11 @@ class Connection:
     #         }
     #     )
 
+    async def close(self) -> None:
+        """Close connection."""
+        await self.websession.aclose()
+        _LOGGER.debug("Connection closed.")
+
     async def get_authorization_code(
         self,
         email: Text,
@@ -382,8 +401,22 @@ class Connection:
             _LOGGER.debug("No email or password for login; unable to login.")
             return
         url = self.get_authorization_code_link(new=True)
-        resp = await self.websession.get(url)
+        resp = await self.websession.get(str(url.update_query({"login_hint": email})))
         html = await resp.text()
+        if resp.history:
+            for item in resp.history:
+                if (
+                    item.status in [301, 302, 303, 304, 305, 306, 307, 308]
+                    and resp.url.host != self.auth_domain.host
+                ):
+                    _LOGGER.debug(
+                        "Detected %s redirect from %s to %s; changing proxy host",
+                        item.status,
+                        item.url.host,
+                        resp.url.host,
+                    )
+                    self.auth_domain = self.auth_domain.with_host(str(resp.url.host))
+                    url = self.get_authorization_code_link()
         soup: BeautifulSoup = BeautifulSoup(html, "html.parser")
         data = get_inputs(soup)
         data["identity"] = email
@@ -391,13 +424,18 @@ class Connection:
         transaction_id: Text = data.get("transaction_id")
         for attempt in range(retry_limit):
             _LOGGER.debug("Attempt #%s", attempt)
-            resp = await self.websession.post(url, data=data)
+            resp = await self.websession.post(str(url), data=data)
             _process_resp(resp)
             if not resp.history:
                 html = await resp.text()
                 if "/mfa/verify" in html:
+                    _LOGGER.debug("Detected MFA request")
                     mfa_resp = await self.websession.get(
-                        "https://auth.tesla.com/oauth2/v3/authorize/mfa/factors",
+                        str(
+                            self.auth_domain.with_path(
+                                "/oauth2/v3/authorize/mfa/factors"
+                            )
+                        ),
                         params={"transaction_id": transaction_id},
                     )
                     _process_resp(mfa_resp)
@@ -415,8 +453,8 @@ class Connection:
                     #         }
                     #     ]
                     # }
-                    mfa_json = await mfa_resp.json()
-                    if len(mfa_json.get("data", [])) > 1:
+                    mfa_json = mfa_resp.json()
+                    if len(mfa_json.get("data", [])) >= 1:
                         factor_id = mfa_json["data"][mfa_device]["id"]
                     if not mfa_code:
                         _LOGGER.debug("No MFA provided")
@@ -425,7 +463,11 @@ class Connection:
                             "MFA Code missing", devices=mfa_json["data"]
                         )
                     mfa_resp = await self.websession.post(
-                        "https://auth.tesla.com/oauth2/v3/authorize/mfa/verify",
+                        str(
+                            self.auth_domain.with_path(
+                                "/oauth2/v3/authorize/mfa/verify"
+                            )
+                        ),
                         json={
                             "transaction_id": transaction_id,
                             "factor_id": factor_id,
@@ -433,7 +475,7 @@ class Connection:
                         },
                     )
                     _process_resp(mfa_resp)
-                    mfa_json = await mfa_resp.json()
+                    mfa_json = mfa_resp.json()
                     if not (
                         mfa_json["data"].get("approved")
                         and mfa_json["data"].get("valid")
@@ -442,13 +484,14 @@ class Connection:
                         raise IncompleteCredentials(
                             "MFA Code invalid", devices=mfa_json["data"]
                         )
-                    resp = await self.websession.post(url, data=data)
+                    resp = await self.websession.post(str(url), data=data)
                     _process_resp(resp)
             await asyncio.sleep(3)
-        if not (resp.history):
+        if not resp.history or not URL(resp.history[-1].url).query.get("code"):
             _LOGGER.debug("Failed to authenticate")
             raise IncompleteCredentials("Unable to login with credentials")
         code_url = URL(resp.history[-1].url)
+        _LOGGER.debug("Found code %s", code_url.query.get("code"))
         return code_url.query.get("code")
 
     def get_authorization_code_link(self, new=False) -> yarl.URL:
@@ -472,7 +515,7 @@ class Connection:
             "scope": "openid email offline_access",
             "state": state,
         }
-        url = yarl.URL("https://auth.tesla.com/oauth2/v3/authorize")
+        url = self.auth_domain.with_path("/oauth2/v3/authorize")
         url = url.update_query(query)
         return url
 
@@ -491,10 +534,10 @@ class Connection:
             "redirect_uri": "https://auth.tesla.com/void/callback",
         }
         auth = await self.websession.post(
-            "https://auth.tesla.com/oauth2/v3/token",
+            str(self.auth_domain.with_path("/oauth2/v3/token")),
             data=oauth,
         )
-        return await auth.json()
+        return auth.json()
 
     async def refresh_access_token(self, refresh_token):
         """Refresh access token from sso."""
@@ -510,10 +553,10 @@ class Connection:
             "scope": "openid email offline_access",
         }
         auth = await self.websession.post(
-            "https://auth.tesla.com/oauth2/v3/token",
+            self.auth_domain.with_path("/oauth2/v3/token"),
             data=oauth,
         )
-        return await auth.json()
+        return auth.json()
 
     async def get_bearer_token(self, access_token):
         """Get bearer token. This is used by the owners API."""
@@ -532,7 +575,7 @@ class Connection:
         auth = await self.websession.post(
             "https://owner-api.teslamotors.com/oauth/token", headers=head, data=oauth
         )
-        return await auth.json()
+        return auth.json()
 
 
 def get_inputs(soup: BeautifulSoup, searchfield=None) -> Dict[str, str]:
@@ -559,7 +602,7 @@ def _process_resp(resp) -> Text:
             _LOGGER.debug("%s: redirected from\n%s", item.method, item.url)
     url = str(resp.request_info.url)
     method = resp.request_info.method
-    status = resp.status
+    status = resp.status_code
     reason = resp.reason
     headers = resp.request_info.headers
     _LOGGER.debug(
